#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


import time

import pytest

import nannos as nn
from nannos import numpy as np

np.random.seed(1984)
Nx = Ny = 2 ** 5

x = np.random.rand(Nx * Ny)


def test_para():
    res = []
    timing = []
    F = [1.1, 1.2, 1.3, 1.4]

    for n_jobs in [1, 2]:

        @nn.parloop(n_jobs=n_jobs)
        def sim(f):
            xa = np.reshape(x, (Nx, Ny))
            eps_pattern = 2 + 1 * xa
            sup = nn.Layer("Superstrate")
            sub = nn.Layer("Substrate")
            ms = nn.Layer("ms", 1)
            pattern = nn.Pattern(eps_pattern)
            ms.add_pattern(pattern)
            sim = nn.Simulation(
                nn.Lattice(([1, 0], [0, 1])), [sup, ms, sub], nn.PlaneWave(f), 50
            )
            R, T = sim.diffraction_efficiencies()
            return R

        t0 = nn.tic()
        res.append(sim(F))
        t1 = nn.toc(t0)
        timing.append(t1)

    speedup = timing[0] / np.array(timing)

    print(f"speedup = {speedup}")

    assert np.allclose(res[0], res[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

from jax.config import config

config.update("jax_enable_x64", True)
from jax import numpy as npj


def null_jax(A, eps=1e-11):
    u, s, vh = npj.linalg.svd(A, full_matrices=False)
    null_mask = s <= eps
    null_space = npj.compress(null_mask, vh, axis=0)
    return npj.transpose(null_space)


def eig_jax(A, eps=1e-11):
    l = npj.linalg.eigvals(A)
    N = A.shape[0]
    v = []
    for j in range(N):
        aj = null_jax(A - l[j] * npj.eye(N), eps=eps)
        v.append(aj)
    return l, npj.conj(npj.array(v)[:, :, 0].T)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


import numpy as np
import pytest

from nannos.formulations.tangent import get_tangent_field

n2 = 7
Nx, Ny = 2 ** n2, 2 ** n2
radius = 0.1
grid = np.ones((Nx, Ny), dtype=float)
x0 = np.linspace(0, 1.0, Nx)
y0 = np.linspace(0, 1.0, Ny)
x, y = np.meshgrid(x0, y0, indexing="ij")
hole = (x - 0.3) ** 2 + (y - 0.3) ** 2 < radius ** 2
square = np.logical_or(np.abs(x + 0.7) < 0.1, np.abs(y + 0.7) < 0.2)
square0 = np.logical_and(x > 0.6, x < 0.8)
square1 = np.logical_and(y > 0.5, y < 0.8)
square = np.logical_and(square1, square0)

grid[hole] = 0
grid[square] = 0


def test_tangent():
    t = get_tangent_field(grid)
    norm = np.linalg.norm(t, axis=0)
    assert np.allclose(norm, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


from . import numpy as np
from .formulations import fft
from .formulations.analytical import fourier_transform_circle
from .formulations.jones import get_jones_field
from .formulations.tangent import get_tangent_field
from .utils import block, get_block, norm

__all__ = ["Simulation"]


class Simulation:
    """Main simulation object.

    Parameters
    ----------
    lattice : :class:`~nannos.Lattice`
        The lattice.
    layers : list
        A list of :class:`~nannos.Layer` objects (the default is ``None``).
    excitation : :class:`~nannos.PlaneWave`
        A plane wave excitation (the default is ``None``).
    nh : int
        Number of Fourier harmonics (the default is ``100``).
    formulation : str
        Formulation type.  (the default is ``'original'``).
        Available formulations are ``'original'``, ``'tangent'`` and ``'jones'``.
    truncation : str
        Truncation method.  (the default is ``'circular'``).
        Available methods are ``'circular'`` and ``'parallelogrammic'``.

    """

    def __init__(
        self,
        lattice,
        layers=[],
        excitation=None,
        nh=100,
        formulation="original",
        truncation="circular",
    ):
        self.lattice = lattice
        self.layers = layers or []
        self.excitation = excitation
        self.nh0 = int(nh)
        if self.nh0 == 1:
            self.nh0 = 2
        self.formulation = formulation
        self.truncation = truncation
        self.harmonics, self.nh = self.lattice.get_harmonics(
            self.nh0, method=self.truncation
        )
        self.omega = 2 * np.pi * self.excitation.frequency
        self.k0para = np.array(self.excitation.wavevector[:2]) * np.sqrt(
            self.layers[0].epsilon * self.layers[0].mu
        )
        r = self.lattice.reciprocal
        self.kx = (
            self.k0para[0]
            + r[0, 0] * self.harmonics[0, :]
            + r[0, 1] * self.harmonics[1, :]
        )
        self.ky = (
            self.k0para[1]
            + r[0, 1] * self.harmonics[0, :]
            + r[1, 1] * self.harmonics[1, :]
        )
        self.Kx = np.diag(self.kx)
        self.Ky = np.diag(self.ky)
        self.IdG = np.eye(self.nh)
        self.ZeroG = np.zeros_like(self.IdG)

        self.a0 = np.zeros(2 * self.nh, dtype=complex)

        # self.a0[0] = self.excitation.amplitude[0]
        # self.a0[self.nh] = self.excitation.amplitude[1]

        self.a0 = []
        for i in range(self.nh * 2):
            if i == 0:
                a0 = self.excitation.amplitude[0]
            elif i == self.nh:
                a0 = self.excitation.amplitude[1]
            else:
                a0 = 0
            self.a0.append(a0)
        self.a0 = np.array(self.a0)

        self.bN = np.zeros(2 * self.nh, dtype=complex)

        self.is_solved = False

        self._layer_names = [l.name for l in self.layers]
        if np.unique(self._layer_names).size != len(self._layer_names):
            raise ValueError("Layers must have different names")

    def _get_layer(self, id):
        if isinstance(id, str):
            if id in self._layer_names:
                for i, l in enumerate(self.layers):
                    if l.name == id:
                        return l, i
            else:
                raise ValueError(f"Unknown layer name {id}")
        elif isinstance(id, int):
            return self.layers[id], id
        else:
            raise ValueError(
                f"Wrong id for layer: {id}. Please use an integrer specifying the layer index or a string for the layer name"
            )

    def solve(self):
        """Solve the grating problem."""
        layers_solved = []
        for layer in self.layers:
            layer = self._build_matrix(layer)
            # layer.solve_eigenproblem(layer.matrix)
            if layer.is_uniform:
                # layer.eigenvectors = np.eye(self.nh*2)
                layer.solve_uniform(self.omega, self.kx, self.ky, self.nh)
            else:
                layer.solve_eigenproblem(layer.matrix)
            layers_solved.append(layer)
        self.layers = layers_solved
        self.is_solved = True

    def get_S_matrix(self, indices=None):
        """Compute the scattering matrix.

        Parameters
        ----------
        indices : list
            Indices ``[i_first,i_last]`` of the first and last layer (the default is ``None``).
            By default get the S matrix of the full stack.

        Returns
        -------
        list
            The scattering matrix ``[[S11, S12], [S21,S22]]``.

        """
        if not self.is_solved:
            self.solve()

        S11 = np.eye(2 * self.nh, dtype=complex)
        S12 = np.zeros_like(S11)
        S21 = np.zeros_like(S11)
        S22 = np.eye(2 * self.nh, dtype=complex)

        if indices is None:
            n_interfaces = len(self.layers) - 1
            stack = range(n_interfaces)
        else:
            stack = range(indices[0], indices[1])

        for i in stack:
            layer, layer_next = self.layers[i], self.layers[i + 1]
            z = layer.thickness or 0
            z_next = layer_next.thickness or 0
            f, f_next = np.diag(phasor(layer.eigenvalues, z)), np.diag(
                phasor(layer_next.eigenvalues, z_next)
            )
            I_ = _build_Imatrix(layer, layer_next)
            I = [[get_block(I_, i, j, 2 * self.nh) for j in range(2)] for i in range(2)]
            A = I[0][0] - f @ S12 @ I[1][0]
            B = np.linalg.inv(A)
            S11 = B @ f @ S11
            S12 = B @ ((f @ S12 @ I[1][1] - I[0][1]) @ f_next)
            S21 = S22 @ I[1][0] @ S11 + S21
            S22 = S22 @ I[1][0] @ S12 + S22 @ I[1][1] @ f_next
        S = [[S11, S12], [S21, S22]]
        if indices is None:
            self.S = S

        return S

    def get_z_poynting_flux(self, layer, an, bn):
        if not self.is_solved:
            self.solve()
        q, phi = layer.eigenvalues, layer.eigenvectors
        A = (layer.Qeps @ phi) @ np.diag(1 / (self.omega * q))
        pa, pb = phi @ an, phi @ bn
        Aa, Ab = A @ an, A @ bn
        cross_term = 0.5 * (np.conj(pb) * Aa - np.conj(Ab) * pa)
        forward_xy = np.real(np.conj(Aa) * pa) + cross_term
        backward_xy = -np.real(np.conj(Ab) * pb) + np.conj(cross_term)
        forward = forward_xy[: self.nh] + forward_xy[self.nh :]
        backward = backward_xy[: self.nh] + backward_xy[self.nh :]
        return forward, backward

    def get_field_fourier(self, layer_index, z=0):
        if not self.is_solved:
            self.solve()

        layer, layer_index = self._get_layer(layer_index)

        ai0, bi0 = self._get_amplitudes(layer_index, translate=False)

        Z = [z] if np.isscalar(z) else z

        fields = []
        for z_ in Z:
            ai, bi = _translate_amplitudes(layer, z_, ai0, bi0)

            ht_fourier = layer.eigenvectors @ (ai + bi)
            hx, hy = ht_fourier[: self.nh], ht_fourier[self.nh :]

            A = (ai - bi) / (self.omega * layer.eigenvalues)
            B = layer.eigenvectors @ A
            et_fourier = layer.Qeps @ B
            ey, ex = -et_fourier[: self.nh], et_fourier[self.nh :]

            hz = (self.kx * ey - self.ky * ex) / self.omega

            ez = (self.ky * hx - self.kx * hy) / self.omega
            if layer.is_uniform:
                ez = ez / layer.epsilon
            else:
                ez = layer.eps_hat_inv @ ez

            fields.append([[ex, ey, ez], [hx, hy, hz]])

        self.fields_fourier = np.array(fields)
        return self.fields_fourier

    def get_ifft_amplitudes(self, amplitudes, shape, axes=(0, 1)):

        amplitudes = np.array(amplitudes)
        # print("amplitudes.shape", amplitudes.shape)
        if len(amplitudes.shape) == 1:
            amplitudes = np.reshape(amplitudes, amplitudes.shape + (1,))

        s = 0
        for i in range(self.nh):
            f = np.zeros(shape + (amplitudes.shape[0],), dtype=complex)
            # print("f.shape", f.shape)
            f[self.harmonics[0, i], self.harmonics[1, i], :] = 1.0
            a = amplitudes[:, i]
            # print("a.shape", a.shape)
            s += a * f

        # print("s.shape", s.shape)
        ft = fft.inverse_fourier_transform(s, axes=axes)
        # print("ft.shape", ft.shape)
        return ft

    def get_field_grid(self, layer_index, z=0, shape=None):
        layer, layer_index = self._get_layer(layer_index)
        shape = shape or layer.patterns[0].epsilon.shape

        fields_fourier = self.get_field_fourier(layer_index, z)

        fe = fields_fourier[:, 0]
        fh = fields_fourier[:, 1]

        E = np.array([self.get_ifft_amplitudes(fe[:, i, :], shape) for i in range(3)])
        H = np.array([self.get_ifft_amplitudes(fh[:, i, :], shape) for i in range(3)])
        return E, H

    def diffraction_efficiencies(self, orders=False):
        """Compute the diffraction efficiencies.

        Parameters
        ----------
        orders : bool
            If ``True``, returns diffracted orders, else returns the sum of
            reflection and transmission for all propagating orders (the default is ``False``).

        Returns
        -------
        tuple
            The reflection and transmission ``R`` and ``T``.

        """
        # if not hasattr(self, "S"):
        # self.get_S_matrix()

        aN, b0 = self._solve_ext()
        fwd_in, bwd_in = self.get_z_poynting_flux(self.layers[0], self.a0, b0)
        fwd_out, bwd_out = self.get_z_poynting_flux(self.layers[-1], aN, self.bN)

        F0 = np.cos(self.excitation.theta) / np.sqrt(
            self.layers[0].epsilon * self.layers[0].mu
        )

        R = np.real(-bwd_in / F0)
        T = np.real(fwd_out / F0)
        if not orders:
            R = np.sum(R)
            T = np.sum(T)
        return R, T

    def get_z_stress_tensor_integral(self, layer_index, z=0):
        if not hasattr(self, "fields_fourier"):
            self.get_field_fourier(layer_index, z=z)
        # ex, ey, ez = self.fields_fourier[0, 0,:,:]
        # hx, hy, hz = self.fields_fourier[0, 1,:,:]

        e = self.fields_fourier[0, 0]
        h = self.fields_fourier[0, 1]
        ex = e[0]
        ey = e[1]
        ez = e[2]

        hx = h[0]
        hy = h[1]
        hz = h[2]

        layer, layer_index = self._get_layer(layer_index)
        dz = (self.ky * hx - self.kx * hy) / self.omega
        if layer.is_uniform:
            dx = ex * layer.epsilon
            dy = ey * layer.epsilon
        else:
            exy = np.hstack((-ey, ex))
            dxy = layer.eps_hat @ exy
            dx = dxy[self.nh :]
            dy = -dxy[: self.nh]

        Tx = np.sum(ex * np.conj(dz) + hx * np.conj(hz), axis=-1).real
        Ty = np.sum(ey * np.conj(dz) + hy * np.conj(hz), axis=-1).real
        Tz = (
            0.5
            * np.sum(
                ez * np.conj(dz)
                + hz * np.conj(hz)
                - ex * np.conj(dx)
                - ey * np.conj(dy)
                - hx * np.conj(hx)
                - hy * np.conj(hy),
                axis=-1,
            ).real
        )
        return Tx, Ty, Tz

    def get_order_index(self, order):
        try:
            len(order) == 2
        except:
            raise ValueError("order must be a tuple of length 2")
        return [k for k, i in enumerate(self.harmonics.T) if np.allclose(i, order)][0]

    def get_order(self, A, order):
        return A[self.get_order_index(order)]

    def _get_toeplitz_matrix(self, u, transverse=False):
        if transverse:
            return [
                [self._get_toeplitz_matrix(u[i, j]) for j in range(2)] for i in range(2)
            ]
        else:
            if self.formulation == "analytical":
                # TODO: analytical FT
                raise NotImplementedError(
                    "Analytical formulation not currently available."
                )
            else:
                uft = fft.fourier_transform(u)
            ix = range(self.nh)
            jx, jy = np.meshgrid(ix, ix, indexing="ij")
            delta = self.harmonics[:, jx] - self.harmonics[:, jy]
            return uft[delta[0, :], delta[1, :]]

    def _build_matrix(self, layer):
        if layer.iscopy:
            layer.matrix = layer.original.matrix
            layer.Kmu = layer.original.Kmu
            layer.eps_hat = layer.original.eps_hat
            layer.eps_hat_inv = layer.original.eps_hat_inv
            layer.mu_hat = layer.original.mu_hat
            layer.mu_hat_inv = layer.original.mu_hat_inv
            layer.Keps = layer.original.Keps
            layer.Peps = layer.original.Peps
            layer.Qeps = layer.original.Qeps
            return layer
        Kx, Ky = self.Kx, self.Ky
        if layer.is_uniform:
            epsilon = layer.epsilon
            mu = layer.mu
        else:
            epsilon = layer.patterns[0].epsilon
            mu = layer.patterns[0].mu
        is_mu_anisotropic = np.shape(mu)[:2] == (3, 3)
        is_epsilon_anisotropic = np.shape(epsilon)[:2] == (3, 3)
        if layer.is_uniform:
            I = np.eye(self.nh)
            if is_epsilon_anisotropic:
                _epsilon = block(
                    [
                        [epsilon[0, 0] * I, epsilon[0, 1] * I],
                        [epsilon[1, 0] * I, epsilon[1, 1] * I],
                    ]
                )
            else:
                _epsilon = epsilon

            if is_mu_anisotropic:
                _mu = block(
                    [[mu[0, 0] * I, mu[0, 1] * I], [mu[1, 0] * I, mu[1, 1] * I]]
                )
            else:
                _mu = mu
            Keps = _build_Kmatrix(1 / epsilon * self.IdG, Ky, -Kx)
            # Pmu = np.eye(self.nh * 2)
            Pmu = block([[mu * self.IdG, self.ZeroG], [self.ZeroG, mu * self.IdG]])

            Qeps = self.omega ** 2 * Pmu - Keps
        else:
            epsilon_zz = epsilon[2, 2] if is_epsilon_anisotropic else epsilon
            mu_zz = mu[2, 2] if is_mu_anisotropic else mu
            # TODO: check if mu or epsilon is homogeneous, no need to compute the Toepliz matrix

            eps_hat = self._get_toeplitz_matrix(epsilon_zz)
            # eps_hat = self._get_toeplitz_matrix(epsilon_zz,ana=False)
            mu_hat = self.IdG  # self._get_toeplitz_matrix(mu_zz)
            eps_hat_inv = np.linalg.inv(eps_hat)
            mu_hat_inv = np.linalg.inv(mu_hat)
            Keps = _build_Kmatrix(eps_hat_inv, Ky, -Kx)
            Kmu = _build_Kmatrix(mu_hat_inv, Kx, Ky)

            if is_epsilon_anisotropic:
                eps_para_hat = self._get_toeplitz_matrix(epsilon, transverse=True)
            else:
                eps_para_hat = [[eps_hat, self.ZeroG], [self.ZeroG, eps_hat]]

            if self.formulation == "original":
                Peps = block(eps_para_hat)
            elif self.formulation == "jones":
                t = get_tangent_field(epsilon_zz, normalize=True)
                # norm_t = norm(t)
                # t /= norm_t.max()
                J = get_jones_field(t)
                # J /= norm(J).max()
                Peps = self._get_Peps(epsilon, eps_para_hat, J, direct=False)
            elif self.formulation == "tangent":
                t = get_tangent_field(epsilon_zz)
                Peps = self._get_Peps(epsilon, eps_para_hat, t)
            else:
                raise ValueError(
                    f"Unknown formulation {self.formulation}. Please choose between 'original', 'tangent' or 'jones'"
                )

            if is_mu_anisotropic:
                mu_para_hat = [
                    [self._get_toeplitz_matrix(mu[i, j]) for j in range(2)]
                    for i in range(2)
                ]

                Pmu = block(mu_para_hat)
            else:
                Pmu = block([[mu_hat, self.ZeroG], [self.ZeroG, mu_hat]])

            # Qeps = self.omega ** 2 * np.eye(self.nh * 2) - Keps
            # matrix = Peps @ Qeps - Kmu
            Qeps = self.omega ** 2 * Pmu - Keps
            matrix = self.omega ** 2 * Peps @ Pmu - (Peps @ Keps + Kmu @ Pmu)

            layer.matrix = matrix
            layer.Kmu = Kmu
            layer.eps_hat = eps_hat
            layer.eps_hat_inv = eps_hat_inv
            layer.mu_hat = mu_hat
            layer.mu_hat_inv = mu_hat_inv
            layer.Keps = Keps
            layer.Peps = Peps

        layer.Qeps = Qeps

        return layer

    def _get_amplitudes(self, layer_index, z=0, translate=True):
        n_interfaces = len(self.layers) - 1
        if layer_index == 0:
            aN, b0 = self._solve_ext()
            ai, bi = self.a0, b0
        elif layer_index == n_interfaces:
            aN, b0 = self._solve_ext()
            ai, bi = aN, self.bN
        else:
            ai, bi = self._solve_int(layer_index)
            layer = self.layers[layer_index]
        if translate:
            ai, bi = _translate_amplitudes(self.layers[layer_index], z, ai, bi)
        return ai, bi

    def _solve_int(self, layer_index):
        n_interfaces = len(self.layers) - 1
        S = self.get_S_matrix(indices=(0, layer_index))
        P = self.get_S_matrix(indices=(layer_index, n_interfaces))
        q = np.linalg.inv(np.eye(self.nh * 2) - np.dot(S[0][1], P[1][0]))
        ai = np.dot(q, np.dot(S[0][0], self.a0))
        bi = np.dot(P[1][0], ai)
        return ai, bi

    def _solve_ext(self):
        if not hasattr(self, "S"):
            self.get_S_matrix()
        aN = self.S[0][0] @ self.a0 + self.S[0][1] @ self.bN
        b0 = self.S[1][0] @ self.a0 + self.S[1][1] @ self.bN
        return aN, b0

    def _get_Peps(self, epsilon, eps_hat, t, direct=False):

        is_epsilon_anisotropic = np.shape(epsilon)[:2] == (3, 3)
        if is_epsilon_anisotropic:
            N = epsilon.shape[2]
            eb = block(epsilon[:2, :2])
            nu = _inv2by2block(eb, N)
            # nu = np.linalg.inv(epsilon[2,2])
            nu1 = np.array(_block2list(nu, N))
            nuhat = self._get_toeplitz_matrix(nu1, transverse=True)

            nuhat_inv = _block2list(np.linalg.inv(block(nuhat)), self.nh)

        else:
            N = epsilon.shape[0]
            nuhat = self._get_toeplitz_matrix(1 / epsilon)
            nuhat_inv = np.linalg.inv((nuhat))
        if direct:
            T = block([[t[1], np.conj(t[0])], [-t[0], np.conj(t[1])]])
            invT = _inv2by2block(T, N)
            if is_epsilon_anisotropic:
                Q = block(
                    [[eps_hat[0][0], nuhat_inv[0][1]], [eps_hat[1][0], nuhat_inv[1][1]]]
                )

            else:
                Q = block([[eps_hat[0][0], self.ZeroG], [self.ZeroG, nuhat_inv]])

            That = block(
                [
                    [self._get_toeplitz_matrix(get_block(T, i, j, N)) for j in range(2)]
                    for i in range(2)
                ]
            )
            invThat = block(
                [
                    [
                        self._get_toeplitz_matrix(get_block(invT, i, j, N))
                        for j in range(2)
                    ]
                    for i in range(2)
                ]
            )
            Peps = That @ Q @ invThat
        else:
            norm_t = norm(t)
            nt2 = np.abs(norm_t) ** 2
            Pxx = t[0] ** 2 / nt2
            Pyy = t[1] ** 2 / nt2
            Pxy = t[0] * np.conj(t[1]) / nt2
            Pyx = t[1] * np.conj(t[0]) / nt2
            Pxx_hat = self._get_toeplitz_matrix(Pxx)
            Pyy_hat = self._get_toeplitz_matrix(Pyy)
            Pxy_hat = self._get_toeplitz_matrix(Pxy)
            Pyx_hat = self._get_toeplitz_matrix(Pyx)

            Pi = block([[Pyy_hat, Pyx_hat], [Pxy_hat, Pxx_hat]])

            # Delta = epsilon - 1 / epsilon
            # Delta_hat = self._get_toeplitz_matrix(Delta)

            eps_para_hat = block(eps_hat)
            if is_epsilon_anisotropic:
                D = eps_para_hat - block(nuhat_inv)
            else:
                D = eps_para_hat - block(
                    [[nuhat_inv, self.ZeroG], [self.ZeroG, nuhat_inv]]
                )

            Peps = eps_para_hat - D @ Pi
        return Peps


def phasor(q, z):
    return np.exp(1j * q * z)


def _build_Kmatrix(u, Kx, Ky):
    return block(
        [
            [Kx @ u @ Kx, Kx @ u @ Ky],
            [Ky @ u @ Kx, Ky @ u @ Ky],
        ]
    )


def _build_Mmatrix(layer):
    phi = layer.eigenvectors
    a = layer.Qeps @ phi @ (np.diag(1 / layer.eigenvalues))
    return block([[a, -a], [phi, phi]])


def _build_Imatrix(layer1, layer2):
    a1 = _build_Mmatrix(layer1)
    a2 = _build_Mmatrix(layer2)
    return np.linalg.inv(a1) @ a2


def _translate_amplitudes(layer, z, ai, bi):
    q = layer.eigenvalues
    aim = ai * phasor(q, z)
    bim = bi * phasor(q, layer.thickness - z)
    return aim, bim


def _block2list(M, N):
    return [[get_block(M, i, j, N) for j in range(2)] for i in range(2)]


def _inv2by2block(T, N):
    M = _block2list(T, N)
    detT = M[0][0] * M[1][1] - M[1][0] * M[0][1]
    return block(
        [
            [M[1][1] / detT, -M[0][1] / detT],
            [-M[1][0] / detT, M[0][0] / detT],
        ]
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

import numpy as np
import pytest

from nannos.layers import Layer

np.random.seed(84)
N = 3


def test_layer():
    l = Layer("test", 1)
    assert l.name == "test"
    assert l.thickness == 1
    assert l.__repr__() == "Layer test"

    matrix = np.random.rand(N, N) + 1j * np.random.rand(N, N)
    w, v = l.solve_eigenproblem(matrix)

    lc = l.copy()
    lc.name = "copy"
    lc.thickness = 2.1
    assert np.allclose(lc.eigenvalues, w)
    assert np.allclose(lc.eigenvectors, v)

    with pytest.raises(ValueError) as excinfo:
        l = Layer("test", -0.1)
    assert "thickness must be positive." == str(excinfo.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


from .. import numpy as np
from ..constants import pi
from ..utils import norm


def get_jones_field(t):
    norm_t = norm(t)
    n = np.array([-t[1], t[0]])
    theta_J = np.arccos(t[0] / norm_t)
    phi_J = pi / 8 * (1 + np.cos(pi * norm_t))
    J = np.exp(1j * theta_J) / norm_t * (t * np.cos(phi_J) + 1j * n * np.sin(phi_J))
    return J



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


__all__ = ["parloop"]


from functools import wraps

from joblib import Parallel, delayed

from . import numpy as np


def parloop(n_jobs=1):
    def deco_parloop(func):
        """
        Decorate a function to parallelize.
        """

        @wraps(func)
        def my_func(*args, **kwargs):
            other_args = args[1:]
            return Parallel(n_jobs=n_jobs)(
                delayed(func)(x, *other_args, **kwargs) for x in args[0]
            )

        return my_func

    return deco_parloop



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

import random

import pytest

random.seed(1984)
Nx = Ny = 4


# @pytest.mark.skip(reason="jax does not work")
def test_grad():
    res = dict()
    dres = dict()

    random.seed(84)
    x = [random.random() for _ in range(Nx * Ny)]
    # for backend in ["jax", "autograd"]:
    for backend in ["autograd"]:
        import nannos as nn

        nn.set_backend(backend)
        from nannos import grad
        from nannos import numpy as np

        def f(x):
            xa = np.reshape(x, (Nx, Ny))
            eps_pattern = 2 + 1 * xa
            sup = nn.Layer("Superstrate")
            sub = nn.Layer("Substrate")
            ms = nn.Layer("ms", 1)
            pattern = nn.Pattern(eps_pattern)
            ms.add_pattern(pattern)
            sim = nn.Simulation(
                nn.Lattice(([1, 0], [0, 1])), [sup, ms, sub], nn.PlaneWave(1.1), 6
            )
            R, T = sim.diffraction_efficiencies()
            return R

        xa = np.array(x)
        y = f(xa)
        res[backend] = y

        dy = grad(f)(xa)
        dres[backend] = dy

        def first_finite_differences(f, x):
            eps = 1e-3
            return np.array(
                [(f(x + eps * v) - f(x - eps * v)) / (2 * eps) for v in np.eye(len(x))]
            )

        dy_fd = first_finite_differences(f, x)
        assert np.allclose(dy, dy_fd)
    # print(res["jax"], res["autograd"])
    # assert np.allclose(res["jax"], res["autograd"])
    # assert np.allclose(dres["jax"], dres["autograd"])
    # nn.set_backend("numpy")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

from .__about__ import __author__, __description__, __version__
from .log import *


def set_backend(backend):
    """Set the numerical backend.

    Parameters
    ----------
    backend : str
        Either ``numpy``, ``autograd`` or ``jax``.


    """

    import importlib
    import sys

    global JAX
    global AUTOGRAD
    if backend == "autograd":
        AUTOGRAD = True
        log.info("Setting autograd backend")
        try:
            del JAX
        except:
            pass
    elif backend == "jax":
        JAX = True
        log.info("Setting jax backend")
        try:
            del AUTOGRAD
        except:
            pass
    elif backend == "numpy":
        try:
            del JAX
        except:
            try:
                del AUTOGRAD
            except:
                pass
            pass
        log.info("Setting numpy backend")
    else:
        raise ValueError(
            f"Unknown backend '{backend}'. Please choose between 'numpy' 'jax' and 'autograd'."
        )

    import nannos

    importlib.reload(nannos)

    its = [s for s in sys.modules.items() if s[0].startswith("nannos")]
    for k, v in its:
        importlib.reload(v)


def get_backend():
    try:
        AUTOGRAD
    except:
        try:
            JAX
            return "jax"
        except:
            return "numpy"
    return "autograd"


try:
    JAX
    from jax.config import config

    config.update("jax_enable_x64", True)
    from jax import grad, numpy

    # TODO: support jax since it is faster than autograd
    # jax does not support numpy.linalg.eig yet
    # for autodif wrt eigenvectors
    # see: https://github.com/google/jax/issues/2748
except:
    try:
        AUTOGRAD
        from autograd import grad, numpy
    except:
        import numpy

        grad = None


from .constants import *
from .excitation import *
from .lattice import *
from .layers import *
from .parallel import *
from .sample import *
from .simulation import *
from .utils import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


"""
Geometry helpers
"""

__all__ = ["shape_mask", "outline_to_mask"]

from shapely.geometry import Polygon

from . import numpy as np

# from https://gist.github.com/perrette/a78f99b76aed54b6babf3597e0b331f8


def outline_to_mask(line, x, y):
    """Create mask from outline contour

    Parameters
    ----------
    line: array-like (N, 2)
    x, y: 1-D grid coordinates (input for meshgrid)

    Returns
    -------
    mask : 2-D boolean array (True inside)

    Examples
    --------
    >>> from shapely.geometry import Point
    >>> poly = Point(0,0).buffer(1)
    >>> x = np.linspace(-5,5,100)
    >>> y = np.linspace(-5,5,100)
    >>> mask = outline_to_mask(poly.boundary, x, y)
    """
    import matplotlib.path as mplp

    mpath = mplp.Path(line)
    X, Y = np.meshgrid(x, y)
    points = np.array((X.flatten(), Y.flatten())).T
    mask = mpath.contains_points(points).reshape(X.shape)
    return mask


def _grid_bbox(x, y):
    dx = dy = 0
    return x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2


def _bbox_to_rect(bbox):
    l, r, b, t = bbox
    return Polygon([(l, b), (r, b), (r, t), (l, t)])


def shape_mask(shp, x, y, m=None):
    """Use recursive sub-division of space and shapely contains method to create a raster mask on a regular grid.

    Parameters
    ----------
    shp : shapely's Polygon (or whatever with a "contains" method and intersects method)
    x, y : 1-D numpy arrays defining a regular grid
    m : mask to fill, optional (will be created otherwise)

    Returns
    -------
    m : boolean 2-D array, True inside shape.

    Examples
    --------
    >>> from shapely.geometry import Point
    >>> poly = Point(0,0).buffer(1)
    >>> x = np.linspace(-5,5,100)
    >>> y = np.linspace(-5,5,100)
    >>> mask = shape_mask(poly, x, y)
    """
    rect = _bbox_to_rect(_grid_bbox(x, y))

    if m is None:
        m = np.zeros((y.size, x.size), dtype=bool)

    if not shp.intersects(rect):
        m[:] = False

    elif shp.contains(rect):
        m[:] = True

    else:
        k, l = m.shape

        if k == 1 and l == 1:
            m[:] = shp.contains(Point(x[0], y[0]))

        elif k == 1:
            m[:, : l // 2] = shape_mask(shp, x[: l // 2], y, m[:, : l // 2])
            m[:, l // 2 :] = shape_mask(shp, x[l // 2 :], y, m[:, l // 2 :])

        elif l == 1:
            m[: k // 2] = shape_mask(shp, x, y[: k // 2], m[: k // 2])
            m[k // 2 :] = shape_mask(shp, x, y[k // 2 :], m[k // 2 :])

        else:
            m[: k // 2, : l // 2] = shape_mask(
                shp, x[: l // 2], y[: k // 2], m[: k // 2, : l // 2]
            )
            m[: k // 2, l // 2 :] = shape_mask(
                shp, x[l // 2 :], y[: k // 2], m[: k // 2, l // 2 :]
            )
            m[k // 2 :, : l // 2] = shape_mask(
                shp, x[: l // 2], y[k // 2 :], m[k // 2 :, : l // 2]
            )
            m[k // 2 :, l // 2 :] = shape_mask(
                shp, x[l // 2 :], y[k // 2 :], m[k // 2 :, l // 2 :]
            )

    return m



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


__all__ = ["adaptive_sampler"]

from . import numpy as np


def adaptive_sampler(f, z0, max_bend=10, max_z_rel=1e-3, max_df=0.05):

    z0 = np.sort(z0)
    zmin = min(z0)
    zmax = max(z0)

    tall = [f(z) for z in z0]
    z = z0.tolist()
    # t = t0.tolist()
    cmax = np.cos(max_bend * np.pi / 180)
    samp = True
    isamp = 0
    if hasattr(tall[0], "__len__") and len(tall[0]) > 0:
        t = [T[0] for T in tall]
        multi_output = True
    else:
        multi_output = False
        t = tall.copy()

    while samp:
        tmin = np.min(t)
        tmax = np.max(t)
        b = []
        for iz in range(len(z) - 2):
            ztmp = z[iz : iz + 3]
            ttmp = t[iz : iz + 3]

            xp, x0, xn = ztmp
            yp, y0, yn = ttmp

            min_dz = max_z_rel * (zmax - zmin)
            min_dt = max_df * (tmax - tmin)

            refx = xn - x0 < min_dz and x0 - xp < min_dz
            refy = abs(y0 - yp) < min_dt and abs(yn - y0) < min_dt

            local_y_max = yp
            if y0 > local_y_max:
                local_y_max = y0
            if yn > local_y_max:
                local_y_max = yn
            local_y_min = yp
            if y0 < local_y_min:
                local_y_min = y0
            if yn < local_y_min:
                local_y_min = yn
            dx0 = (x0 - xp) / (xn - xp)
            dx1 = (xn - x0) / (xn - xp)
            dy0 = (y0 - yp) / (local_y_max - local_y_min)
            dy1 = (yn - y0) / (local_y_max - local_y_min)

            bend = (dx0 * dx1 + dy0 * dy1) / np.sqrt(
                (dx0 * dx0 + dy0 * dy0) * (dx1 * dx1 + dy1 * dy1)
            )

            bending = (bend) < cmax or dx1 > 3 * dx0 or dx0 > 3 * dx1

            b.append(bending)
            if bending and not refx and not refy:
                seg = []
                if x0 - xp < min_dz:
                    isegment = 1
                    seg.append(isegment)
                if xn - x0 < min_dz:
                    isegment = 0
                    seg.append(isegment)
                if x0 - xp > xn - x0:
                    isegment = 0
                else:
                    isegment = 1
                seg.append(isegment)
                seg = np.unique(seg)

                for isegment in seg:
                    # isegment = np.random.randint(2)
                    znew = 0.5 * sum(ztmp[isegment : isegment + 2])
                    if znew not in z:
                        z.append(znew)
                        tnew = f(znew)
                        if multi_output:
                            t.append(tnew[0])
                            tall.append(tnew)
                        else:
                            t.append(tnew)

        z1 = np.array(z)
        t1 = np.array(t)
        ind = np.argsort(z1)
        z = z1[ind].tolist()
        t = t1[ind].tolist()
        if multi_output:
            tall = np.array(tall)[ind].tolist()
        samp = np.any(b)

        isamp += 1
        if isamp > 100:
            break

    tout = tall if multi_output else t
    return np.array(z), np.array(tout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


from . import numpy as np
from .constants import pi

__all__ = ["Lattice"]


class Lattice:
    """A lattice object.

    Parameters
    ----------
    basis_vectors : tuple
        The lattice vectors :math:`((u_x,u_y),(v_x,v_y))`.


    """

    def __init__(self, basis_vectors):
        self.basis_vectors = basis_vectors

    @property
    def area(self):
        v = self.basis_vectors
        return np.linalg.norm(np.cross(v[0], v[1]))

    @property
    def matrix(self):
        """Basis matrix.

        Returns
        -------
        array like
            Matrix containing the lattice vectors (L1,L2).

        """
        return np.array(self.basis_vectors).T

    @property
    def reciprocal(self):
        """Reciprocal matrix.

        Returns
        -------
        array like
            Matrix containing the lattice vectors (K1,K2) in reciprocal space.

        """
        return 2 * pi * np.linalg.inv(self.matrix).T

    def get_harmonics(self, nh, method="circular"):
        """Short summary.

        Parameters
        ----------
        nh : int
            Number of harmonics.
        method : str
            The truncation method, available values are "circular" and "parallelogrammic"
            (the default is "circular").

        Returns
        -------
        type
            Description of returned object.

        """
        if not int(nh) == nh:
            raise ValueError("nh must be integer.")
        if method == "circular":
            return _circular_truncation(nh, self.reciprocal)
        elif method == "parallelogrammic":
            return _parallelogramic_truncation(nh, self.reciprocal)
        else:
            raise ValueError(
                f"Unknown truncation method '{method}', please choose between 'circular' or 'parallelogrammic'."
            )


def _parallelogramic_truncation(nh, Lk):
    u = [np.linalg.norm(l) for l in Lk]
    udot = np.dot(Lk[0], Lk[1])

    NGroot = int(np.sqrt(nh))
    if np.mod(NGroot, 2) == 0:
        NGroot -= 1

    M = NGroot // 2

    xG = range(-M, NGroot - M)
    G = np.meshgrid(xG, xG, indexing="ij")
    G = [g.flatten() for g in G]

    Gl2 = G[0] ** 2 * u[0] ** 2 + G[1] ** 2 * u[0] ** 2 + 2 * G[0] * G[1] * udot
    jsort = np.argsort(Gl2)
    Gsorted = [g[jsort] for g in G]

    nh = NGroot ** 2
    G = np.array(Gsorted, dtype=int)[:, :nh]

    return G, nh


def _circular_truncation(nh, Lk):
    u = [np.linalg.norm(l) for l in Lk]
    udot = np.dot(Lk[0], Lk[1])
    ucross = Lk[0][0] * Lk[1][1] - Lk[0][1] * Lk[1][0]

    circ_area = nh * np.abs(ucross)
    circ_radius = np.sqrt(circ_area / pi) + u[0] + u[1]

    u_extent = [
        1 + int(circ_radius / (q * np.sqrt(1.0 - udot ** 2 / (u[0] * u[1]) ** 2)))
        for q in u
    ]

    xG, yG = [np.arange(-q, q + 1) for q in u_extent]
    G = np.meshgrid(xG, yG, indexing="ij")
    G = [g.flatten() for g in G]

    Gl2 = G[0] ** 2 * u[0] ** 2 + G[1] ** 2 * u[0] ** 2 + 2 * G[0] * G[1] * udot
    jsort = np.argsort(Gl2)
    Gsorted = [g[jsort] for g in G]
    Gl2 = Gl2[jsort]

    nGtmp = (2 * u_extent[0] + 1) * (2 * u_extent[1] + 1)
    if nh < nGtmp:
        nGtmp = nh

    tol = 1e-10 * max(u[0] ** 2, u[1] ** 2)
    for i in np.arange(nGtmp - 1, -1, -1):
        if np.abs(Gl2[i] - Gl2[i - 1]) > tol:
            break
    nh = i

    G = np.array(Gsorted, dtype=int)[:, :nh]

    return G, nh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


__all__ = ["Layer", "Pattern"]

from copy import copy

from . import get_backend
from . import numpy as np
from .simulation import block


class Layer:
    """A layer object.

    Parameters
    ----------
    name : str
        Name of the layer (the default is "layer").
    thickness : float
        Thickness of the layer (the default is 0). Must be positive.
    epsilon : complex
        Permittivity `epsilon` (the default is 1).
    mu : complex
        Permeability `mu` (the default is 1).


    """

    def __init__(self, name="layer", thickness=0, epsilon=1, mu=1):
        if thickness is not None:
            if thickness < 0:
                raise ValueError("thickness must be positive.")
        self.name = name
        self.thickness = thickness
        self.epsilon = epsilon
        self.mu = mu
        self.iscopy = False
        self.original = self
        self.patterns = []

        if self.mu != 1:
            # TODO: case with magnetic materials
            raise NotImplementedError(
                "Permeability different than unity not yet implemented."
            )

    def __repr__(self):
        return f"Layer {self.name}"

    def solve_uniform(self, omega, kx, ky, nh):
        """Solve for eigenmodes and eigenvalues of a uniform layer.

        Parameters
        ----------
        omega : float
            Pulsation.
        kx : array_like
            Transverse wavenumber x component.
        ky : array_like
            Transverse wavenumber y component.
        nh : int
            Number of harmonics.

        Returns
        -------
        tuple
            `(q,psi)`, with eigenvalues `q` and eigenvectors `psi`.

        """
        epsilon = self.epsilon
        mu = self.mu

        is_mu_anisotropic = np.shape(mu)[:2] == (3, 3)
        is_epsilon_anisotropic = np.shape(epsilon)[:2] == (3, 3)

        if is_mu_anisotropic or is_epsilon_anisotropic:
            # TODO: anisotropic uniform layer
            raise NotImplementedError("Uniform layer material must be isotropic")

        IdG = np.eye(2 * nh)

        I = np.eye(nh)
        if is_epsilon_anisotropic:
            _epsilon = block(
                [
                    [epsilon[0, 0] * I, epsilon[0, 1] * I],
                    [epsilon[1, 0] * I, epsilon[1, 1] * I],
                ]
            )
        else:
            _epsilon = epsilon

        if is_mu_anisotropic:
            _mu = block([[mu[0, 0] * I, mu[0, 1] * I], [mu[1, 0] * I, mu[1, 1] * I]])
        else:
            _mu = mu

        q = (
            np.array(
                _epsilon * _mu * omega ** 2 - kx ** 2 - ky ** 2,
                dtype=complex,
            )
            ** 0.5
        )
        q = np.where(np.imag(q) < 0.0, -q, q)
        self.eigenvalues = np.concatenate((q, q))
        self.eigenvectors = np.eye(2 * len(kx))
        return self.eigenvalues, self.eigenvectors

    def solve_eigenproblem(self, matrix):
        """Solve the eigenproblem for a patterned layer.

        Parameters
        ----------
        matrix : array_like
            The matrix which we search for eigenvalues and eigenvectors.

        Returns
        -------
        tuple
            `(q,psi)`, with eigenvalues `q` and eigenvectors `psi`.

        """
        if self.iscopy:
            self.eigenvalues, self.eigenvectors = (
                self.original.eigenvalues,
                self.original.eigenvectors,
            )

        else:
            # FIXME: This gets slow because of the implementation
            # workaround to cpmpute eigenvalues with jax
            # TODO: implement custom autodiff rules for the evp with jax
            if get_backend() == "jax":
                from ._jax_eig_workaround import eig_jax

                eig_func = eig_jax
            else:
                eig_func = np.linalg.eig

            # eig_func = np.linalg.eig

            w, v = eig_func(matrix)
            q = w ** 0.5
            q = np.where(np.imag(q) < 0.0, -q, q)
            self.eigenvalues, self.eigenvectors = q, v
        return self.eigenvalues, self.eigenvectors

    def copy(self, name=None):
        """Copy a layer.

        Returns
        -------
        nannos.Layer
            A copy of the layer.

        """
        cp = copy(self)
        cp.iscopy = True
        cp.original = self
        if name is None:
            cp.name += " (copy)"
        else:
            cp.name = name
        return cp

    def add_pattern(self, pattern):
        """Add a pattern to the layer.

        Parameters
        ----------
        pattern : :class:`~nannos.Pattern`
            The pattern defined as a 2d grid on the unit cell.

        """
        self.patterns.append(pattern)

    @property
    def is_uniform(self):
        """Check if layer is uniform.

        Returns
        -------
        bool
            ``True`` if the layer is uniform, ``False`` if not.

        """
        return len(self.patterns) == 0


class Pattern:
    """A pattern object.

    Parameters
    ----------
    epsilon : array_like
        Permittivity `epsilon` (the default is 1).
    mu : array_like
        Permeability `mu` (the default is 1).
    name : str
        Name of the pattern (the default is "pattern").
    grid : array_like
        A 2d grid on which the pattern is defined (the default is None).



    """

    def __init__(self, epsilon=1, mu=1, name="pattern", grid=None):
        self.name = name
        self.grid = grid
        self.epsilon = epsilon
        self.mu = mu


#
#
#
# def eig2by2(M):
#     tr = M[0, 0] + M[1,1]
#     det = M[0, 0] * M[1,1] - M[1, 0] + M[0,1]
#     l0 = 0.5 * (tr - (tr ** 2 - 4 * det ) ** 0.5)
#     l1 = 0.5 * (tr + (tr ** 2 - 4 * det ) ** 0.5)
#     v0 = [-M[0, 1], M[0, 0] - l0]
#     v1 = [M[1, 1] - l1, -M[1, 0]]
#
#     eigenvalues = np.array([l0, l1])
#     # eigenvectors = np.array([v0, v1])
#     if  M[0, 0] - l0 == 0 or  M[1, 1] - l1 ==0:
#         eigenvectors = np.eye(2)
#     else:
#         eigenvectors = np.array([np.array(v)/(v[0]**2 + v[1]**2)**0.5 for v in [v0, v1]]).T
#
#     return eigenvalues, eigenvectors



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


import numpy as npo

from .. import numpy as np
from ..utils import filter, norm


def _normalize(x, n):
    with npo.errstate(invalid="ignore"):
        f = x / (n)
    return np.where(n == 0.0, 0, f)


def _ft_filt(x, expo):
    with npo.errstate(divide="ignore"):
        f = 1 / (x ** expo)
    return np.where(x == 0.0, 0, f)


def get_tangent_field(grid, normalize=True, alt=False, rfilt=4, expo=0.5):

    Nx, Ny = grid.shape
    xf = filter(grid, rfilt)
    v = np.gradient(grid)
    vf = np.gradient(xf)
    norma = norm(v)
    N = np.array(v) * np.abs(np.array(vf))
    N = _normalize(N, norma)

    fx = np.fft.fftfreq(Nx)
    fy = np.fft.fftfreq(Ny)
    Fx, Fy = np.meshgrid(fx, fy)

    ghat = _ft_filt(Fx ** 2 + Fy ** 2, expo=expo)
    Nhat = np.fft.fft2(N)
    Nstar = np.real(np.fft.ifft2(Nhat * ghat))
    if normalize:
        norm_Nstar = norm(Nstar)
        Nstar = _normalize(Nstar, norm_Nstar)

    if alt:
        return np.array([Nstar[1], Nstar[0]])
    else:
        return np.array([-Nstar[1], Nstar[0]])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


import nlopt
from autograd import grad
from scipy.optimize import minimize

from . import numpy as np
from .utils import filter


def project(x, beta=1, thres=0.5):
    return ((np.tanh(thres * beta)) + np.tanh(beta * (x - thres))) / (
        np.tanh(thres * beta) + (np.tanh((1 - thres) * beta))
    )


class StopFunError(Exception):
    """Raised when stop value reached"""

    pass


class StopFun:
    def __init__(self, fun, stopval=None):
        self.fun_in = fun
        self.stopval = stopval

    def fun(self, x, *args):
        self.f = self.fun_in(x, *args)
        if self.stopval is not None:
            if self.f < self.stopval:
                raise StopFunError("Stop value reached.")
        self.x = x
        return self.f


class TopologyOptimizer:
    def __init__(
        self,
        fun,
        x0,
        method="scipy",
        threshold=(0, 8),
        maxiter=10,
        stopval=None,
        args=None,
        callback=None,
        options={},
    ):
        self.fun = fun
        self.x0 = x0
        self.nvar = len(x0)
        self.method = method
        self.threshold = threshold
        self.maxiter = maxiter
        self.stopval = stopval
        self.args = args
        self.callback = callback
        self.options = options
        self.grad_fun = grad(fun)

    def min_function(self):
        f = self.fun(x)

    def minimize(self):

        print("#################################################")
        print(f"Topology optimization with {self.nvar} variables")
        print("#################################################")
        print("")
        x0 = self.x0
        for iopt in range(*self.threshold):
            print(f"global iteration {iopt}")
            print("-----------------------")

            proj_level = 2 ** iopt
            args = list(self.args)
            args[1] = proj_level
            args = tuple(args)
            if self.method == "scipy":

                def fun_scipy(x, *args):
                    y = fun(x, *args)
                    print(f"current value = {y}")
                    if self.callback is not None:
                        self.callback(x, y, *args)
                    return y

                bounds = [(0, 1) for _ in range(self.nvar)]
                options = {"maxiter": self.maxiter}
                if self.options is not None:
                    options.update(self.options)

                stop = StopFun(fun=fun_scipy, stopval=self.stopval)
                try:
                    opt = minimize(
                        stop.fun,
                        x0,
                        args=args,
                        method="L-BFGS-B",
                        jac=self.grad_fun,
                        bounds=bounds,
                        options=options,
                    )
                    xopt = opt.x
                    fopt = opt.fun
                except StopFunError as e:
                    print(e)
                    xopt = stop.x
                    fopt = stop.f

            else:

                def fun_nlopt(x, gradn):
                    gradn[:] = self.grad_fun(x, *args)
                    y = self.fun(x, *args)
                    print(f"current value = {y}")
                    if self.callback is not None:
                        self.callback(x, y, *args)
                    return y

                lb = np.zeros(self.nvar, dtype=float)
                ub = np.ones(self.nvar, dtype=float)

                opt = nlopt.opt(nlopt.LD_MMA, self.nvar)
                opt.set_lower_bounds(lb)
                opt.set_upper_bounds(ub)

                # opt.set_ftol_rel(1e-16)
                # opt.set_xtol_rel(1e-16)
                if self.stopval is not None:
                    opt.set_stopval(self.stopval)
                if self.maxiter is not None:
                    opt.set_maxeval(self.maxiter)

                opt.set_min_objective(fun_nlopt)
                xopt = opt.optimize(x0)
                fopt = opt.last_optimum_value()
            x0 = xopt
        return xopt, fopt



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


import numpy as np
import pytest

from nannos.lattice import *
from nannos.lattice import Lattice


def test_lattice():
    l = Lattice(((1, 2), (3, 4)))
    assert np.allclose(np.array([[1, 3], [2, 4]]), l.matrix)
    assert np.allclose(
        np.array([[-12.56637061, 6.28318531], [9.42477796, -3.14159265]]), l.reciprocal
    )


def test_truncate():
    l = Lattice(((1, 2), (3, 4)))
    g, nh = l.get_harmonics(100)
    g, nh = l.get_harmonics(100, method="parallelogrammic")
    with pytest.raises(ValueError) as excinfo:
        l.get_harmonics(100, method="unknown")
    assert "Unknown truncation method" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        l.get_harmonics(1.2)
    assert "nh must be integer." == str(excinfo.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

try:
    # Python 3.8
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

try:
    data = metadata.metadata("nannos")
    __version__ = metadata.version("nannos")
    __author__ = data.get("author")
    __description__ = data.get("summary")
except Exception:
    __version__ = "unknown"
    __author__ = "unknown"
    __description__ = "unknown"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

from . import numpy as np

__all__ = ["PlaneWave"]


class PlaneWave:
    """A plane wave object.

    Parameters
    ----------
    frequency : float
        Frequency (the default is 1).
    angles : tuple
        Incidence angles :math:`(\\theta,\phi,\psi)` (the default is (0, 0, 0)).
        :math:`\\theta`: polar angle,
        :math:`\phi`: azimuthal angle,
        :math:`\psi`: polarization angle.


    """

    def __init__(self, frequency=1, angles=(0, 0, 0)):
        self.frequency = frequency
        self.angles = angles
        self.theta, self.phi, self.psi = angles

        k0 = 2 * np.pi * frequency
        self.wavenumber = k0
        self.wavevector = k0 * np.array(
            (
                np.sin(self.theta) * np.cos(self.phi),
                np.sin(self.theta) * np.sin(self.phi),
                np.cos(self.theta),
            )
        )
        cpsi = np.cos(self.psi)  # p
        spsi = np.sin(self.psi)  # s
        cx = -spsi * np.cos(self.theta) * np.cos(self.phi) - cpsi * np.sin(self.phi)
        cy = -spsi * np.cos(self.theta) * np.sin(self.phi) + cpsi * np.cos(self.phi)
        cz = -cpsi * np.sin(self.theta)
        self.amplitude = [cx, cy, cz]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


from .. import numpy as np


def fourier_transform(u):
    uft = np.fft.fft2(u)
    nx, ny = np.shape(uft)[:2]
    return uft / (nx * ny)


def inverse_fourier_transform(uft, shape=None, axes=(-2, -1)):
    u = np.fft.ifft2(uft, s=shape, axes=axes)
    nx, ny = np.shape(u)[:2]
    return u * (nx * ny)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

import pytest


def test_backend():
    import nannos as nn

    assert nn.get_backend() == "numpy"

    nn.set_backend("autograd")
    assert nn.numpy.__name__ == "autograd.numpy"
    assert nn.get_backend() == "autograd"
    # nn.set_backend("jax")
    # assert nn.numpy.__name__ == "jax.numpy"
    # assert nn.get_backend() == "jax"
    with pytest.raises(ValueError) as excinfo:
        nn.set_backend("fake")
    assert "Unknown backend" in str(excinfo.value)
    nn.set_backend("numpy")
    assert nn.numpy.__name__ == "numpy"
    assert nn.get_backend() == "numpy"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


__all__ = ["set_log_level", "log"]


import logging

from colorlog import ColoredFormatter

LEVELS = dict(
    DEBUG=logging.DEBUG,
    INFO=logging.INFO,
    WARNING=logging.WARNING,
    ERROR=logging.ERROR,
    CRITICAL=logging.CRITICAL,
)


def set_log_level(level="INFO"):
    """Sets the log level.

    Parameters
    ----------
    level : str
        The verbosity level.
        Valid values are ``DEBUG``, ``INFO``, ``WARNING``, ``ERROR`` or ``CRITICAL``
        (the default is ``INFO``).

    """

    global log
    LOG_LEVEL = LEVELS[level]
    LOGFORMAT = (
        "  %(log_color)s%(levelname)-8s%(reset)s | %(log_color)s%(message)s%(reset)s"
    )

    logging.root.setLevel(LOG_LEVEL)
    formatter = ColoredFormatter(LOGFORMAT)
    stream = logging.StreamHandler()
    stream.setLevel(LOG_LEVEL)
    stream.setFormatter(formatter)
    log = logging.getLogger("pythonConfig")
    log.setLevel(LOG_LEVEL)
    # if not log.hasHandlers():
    [log.removeHandler(h) for h in log.handlers]
    log.addHandler(stream)


set_log_level("INFO")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

from IPython import get_ipython

from .helpers import *
from .jupyter import VersionTable
from .time import *

_IP = get_ipython()
if _IP is not None:
    _IP.register_magics(VersionTable)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


from .. import numpy as np


def next_power_of_2(x):
    return 1 if x == 0 else int(2 ** np.ceil(np.log2(x)))


def norm(v):
    ## avoid division by 0
    eps = np.finfo(float).eps
    return np.sqrt(eps + np.power(v[0], 2) + np.power(v[1], 2))


def block(a):
    l1 = np.hstack([a[0][0], a[0][1]])
    l2 = np.hstack([a[1][0], a[1][1]])
    return np.vstack([l1, l2])


def get_block(M, i, j, n):
    return M[i * n : (i + 1) * n, j * n : (j + 1) * n]


def filter(x, rfilt):
    if rfilt == 0:
        return x
    else:
        Nx = x.shape[0]
        # First a 1-D  Gaussian
        # t = np.linspace(0, Nx-1, Nx)
        t = np.linspace(-Nx / 2, Nx / 2, Nx)
        bump = np.exp(-(t ** 2) / rfilt ** 2)
        bump /= np.trapz(bump)  # normalize the integral to 1

        # make a 2-D kernel out of it
        kernel = bump[:, np.newaxis] * bump[np.newaxis, :]
        kernel_ft = np.fft.fft2(kernel, s=x.shape[:2], axes=(0, 1))

        # convolve
        img_ft = np.fft.fft2(x, axes=(0, 1))
        img2_ft = kernel_ft * img_ft
        out = np.real(np.fft.ifft2(img2_ft, axes=(0, 1)))
        return np.fft.fftshift(out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


# from nannos import numpy as np
import numpy as np
import pytest

from nannos import Lattice, Layer, Pattern, PlaneWave, Simulation, pi

nh = 51
L1 = [1.0, 0]
L2 = [0, 1.0]
freq = 1.1
theta = 30.0 * pi / 180
phi = 30.0 * pi / 180
psi = 30.0 * pi / 180


pw = PlaneWave(
    frequency=freq, angles=(theta * pi / 180, phi * pi / 180, psi * pi / 180)
)

Nx = 2 ** 9
Ny = 2 ** 9

eps_pattern = 4.0
eps_hole = 1.0
mu_pattern = 1.0
mu_hole = 1.0

h = 2

radius = 0.25
x0 = np.linspace(0, 1.0, Nx)
y0 = np.linspace(0, 1.0, Ny)
x, y = np.meshgrid(x0, y0, indexing="ij")
hole = (x - 0.5) ** 2 + (y - 0.5) ** 2 < radius ** 2

lattice = Lattice((L1, L2))
sup = Layer("Superstrate", epsilon=1, mu=1)
sub = Layer("Substrate", epsilon=1, mu=1)


ids = np.ones((Nx, Ny), dtype=float)
zs = np.zeros_like(ids)


eps = eps_pattern, eps_hole
mu = mu_pattern, mu_hole


def build_pattern(eps, mu, anisotropic=False):
    eps_pattern, eps_hole = eps
    mu_pattern, mu_hole = mu

    if anisotropic:
        exx = ids * eps_pattern
        eyy = ids * eps_pattern * 3
        ezz = ids * eps_pattern * 1
        exx[hole] = eps_hole
        eyy[hole] = eps_hole
        ezz[hole] = eps_hole
        epsgrid = np.array([[exx, zs, zs], [zs, eyy, zs], [zs, zs, ezz]])

        mxx = ids * mu_pattern
        myy = ids * mu_pattern
        mzz = ids * mu_pattern
        mxx[hole] = mu_hole
        myy[hole] = mu_hole
        mzz[hole] = mu_hole
        mugrid = np.array([[mxx, zs, zs], [zs, mxx, zs], [zs, zs, mzz]])
    else:
        epsgrid = ids * eps_pattern
        epsgrid[hole] = eps_hole
        mugrid = ids * mu_pattern
        mugrid[hole] = mu_hole

    return epsgrid, mugrid


@pytest.mark.parametrize("freq", [0.7, 1.1])
@pytest.mark.parametrize("theta", [0, 30])
@pytest.mark.parametrize("phi", [0, 30])
@pytest.mark.parametrize("psi", [0, 30])
def test_uniform(freq, theta, phi, psi):
    pw = PlaneWave(
        frequency=freq, angles=(theta * pi / 180, phi * pi / 180, psi * pi / 180)
    )
    # eps = np.diag([2, 3, 4])
    # mu = np.diag([5, 6, 7])
    eps = 4
    mu = 1
    eps_sup, eps_sub = 1.0, 1.0
    mu_sup, mu_sub = 1, 1

    sup = Layer("Superstrate", epsilon=eps_sup, mu=mu_sup)
    sub = Layer("Substrate", epsilon=eps_sub, mu=mu_sub)

    un = Layer("Uniform", h, epsilon=eps, mu=mu)
    sim = Simulation(lattice, [sup, un, sub], pw, nh=5)
    R, T = sim.diffraction_efficiencies()
    B = R + T
    assert np.allclose(B, 1)

    # sup = Layer("Superstrate", epsilon=mu_sup, mu=eps_sup)
    # sub = Layer("Substrate", epsilon=mu_sub, mu=eps_sub)
    #
    # un = Layer("Uniform", h, epsilon=mu, mu=eps)
    # sim = Simulation(lattice, [sup, un, sub], pw, nh)
    # Rdual, Tdual = sim.diffraction_efficiencies()
    # assert np.allclose(Rdual + Tdual, 1)
    #
    # print(R)
    # print(Rdual)
    # print(T)
    # print(Tdual)
    #
    # assert np.allclose(R, Rdual, rtol=1e-3)
    # assert np.allclose(T, Tdual, rtol=1e-3)


def hole_array(epsgrid, mugrid, pw, nh=nh, formulation="original"):
    pattern = Pattern(epsgrid, mugrid)
    st = Layer("Structured", h)
    st.add_pattern(pattern)
    sim = Simulation(lattice, [sup, st, sub], pw, nh, formulation=formulation)
    return sim


formulations = ["original", "tangent", "jones"]


@pytest.mark.parametrize("freq", [0.7, 1.1])
@pytest.mark.parametrize("theta", [0, 30])
@pytest.mark.parametrize("phi", [0, 30])
@pytest.mark.parametrize("psi", [0, 30])
@pytest.mark.parametrize("formulation", formulations)
def test_fft(freq, theta, phi, psi, formulation):
    pw = PlaneWave(
        frequency=freq, angles=(theta * pi / 180, phi * pi / 180, psi * pi / 180)
    )

    epsgrid, mugrid = build_pattern(eps, mu, anisotropic=False)
    sim = hole_array(epsgrid, mugrid, pw, formulation=formulation)
    R, T = sim.diffraction_efficiencies()
    B = R + T

    print(">>> formulation = ", formulation)
    print("T = ", T)
    print("R = ", R)
    print("R + T = ", B)
    assert np.allclose(B, 1, atol=1e-1)

    a, b = sim._get_amplitudes(1, z=0.1)
    field_fourier = sim.get_field_fourier(1, z=0.1)

    # epsgrid, mugrid = build_pattern(eps, mu, anisotropic=True)
    # simu_aniso = hole_array(epsgrid, mugrid, pw, formulation=formulation)
    # Raniso, Taniso = simu_aniso.diffraction_efficiencies()
    # Baniso = Raniso + Taniso
    # # assert np.allclose(Baniso, 1,atol=1e-3)
    #
    # print(">>> formulation (anisotropic)= ", formulation)
    # print("T = ", Taniso)
    # print("R = ", Raniso)
    # print("R + T = ", Baniso)
    # assert np.allclose(Baniso, 1,atol=1e-1)
    # assert np.allclose(R, Raniso)
    # assert np.allclose(T, Taniso)
    return R, T, sim


# for f in formulations:
#
#     R,T,sim = test_fft(1.7, 0, 0, 0,f)
#     print("**********************************8")
#     #
#     # print(">>> formulation = ", f)
#     # print("T = ", (T))
#     # print("R = ", (R))
#     # B = (R + T)
#     # print("R + T = ", B)
# #



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


from scipy import special as sp

from nannos import numpy as np
from nannos import pi


def jinc(x):
    with np.errstate(invalid="ignore"):
        f = sp.j1(x) / x
    return np.where(x == 0.0, 0.5, f)


def fourier_transform_circle(radius, Nx, Ny):
    dx = 1 / (Nx - 1)
    dy = 1 / (Ny - 1)
    fx = np.fft.fftfreq(Nx, d=dx)
    fy = np.fft.fftfreq(Ny, d=dy)
    kx, ky = np.meshgrid(fx, fy, indexing="ij")
    r = (kx ** 2 + ky ** 2) ** 0.5
    ft = 2 * pi * radius ** 2 * jinc(2 * pi * r * radius) + 0j
    ft *= np.exp(-2 * np.pi * 1j * (0.5 * kx + 0.5 * ky))
    return ft


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    plt.ion()

    plt.close("all")

    Nx = 2 ** 6
    Ny = Nx

    radius = 0.4
    x0 = np.linspace(0, 1, Nx)
    y0 = np.linspace(0, 1, Ny)
    x, y = np.meshgrid(x0, y0, indexing="ij")

    r = ((x - 0.5) ** 2 + (y - 0.5) ** 2) ** 0.5
    hole = r < radius
    from nannos.formulations.fft import fourier_transform

    eb = 4
    eh = 1

    # ids = np.zeros((Nx, Ny), dtype=float)
    ids = np.ones((Nx, Ny), dtype=float) * eb
    ids[hole] = eh

    dx = x0[-1] - x0[0]
    dy = y0[-1] - y0[0]
    d = x0[1] - x0[0]

    # ft = np.fft.fft2(ids)
    ft = fourier_transform(ids)

    # ft = np.fft.fftshift(ft) * (d ** 2)
    # nx, ny = np.shape(ft)
    # ft /= (nx * ny)

    #
    # plt.figure()
    # plt.imshow(ids)
    # plt.colorbar()

    plt.figure()
    plt.imshow(np.real(ft))
    plt.colorbar()
    plt.title("fft")

    f0 = np.fft.fftfreq(Nx, d=d)
    f1 = np.fft.fftshift(f0)

    ana = fourier_transform_circle(radius, Nx, Nx)
    bg = np.zeros_like(ana)
    bg[0, 0] = eb
    ana = bg + (eh - eb) * ana
    # ana = np.fft.fftshift(ana)
    plt.figure()
    plt.imshow(np.real(ana))
    plt.colorbar()
    plt.title("analytical")

    n = 0  # int(Nx / 2)

    plt.figure()
    plt.plot(np.real(ana[:, n]), "o--", label="analytical Re")
    plt.plot(np.real(ft[:, n]), "s-", label="fft Re")
    plt.plot(np.imag(ana[:, n]), "o--", label="analytical Im")
    plt.plot(np.imag(ft[:, n]), "s-", label="fft Im")
    plt.legend()

    plt.figure()
    plt.imshow(np.log10(np.abs(ft - ana)))
    plt.colorbar()
    plt.title("error")

    xs

    n1 = 3
    q = (np.abs(ana - ft) ** 2)[n - n1 : n + n1, n - n1 : n + n1]

    assert np.mean(np.abs(ana - ft) ** 2) < 1e-3
    t = 1e-2
    cond = np.abs(ana) > t
    q = np.abs(ana[cond] - ft[cond]) ** 2

    assert np.mean(q) < 1e-2

    #
    # if ana:
    #     Nx, Ny = u.shape
    #     eb = 4
    #     eh = 1
    #     uft = fourier_transform_circle(0.25, Nx, Ny)
    #     bg = np.zeros_like(uft)
    #     bg[0, 0] = eb
    #     uft = bg + (eh - eb) * uft
